r"""Contain the base class to implement a dataset loader object."""

from __future__ import annotations

__all__ = ["BaseDatasetLoader", "is_dataset_loader_config", "setup_dataset_loader"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    from iden.dataset import BaseDataset

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseDatasetLoader(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a dataset loader.

    A dataset loader object allows to load a ``BaseDataset`` object from
    its Uniform Resource Identifier (URI).

    Example usage:

    ```pycon
    >>> from iden.dataset.loader import VanillaDatasetLoader
    >>> loader = VanillaDatasetLoader()
    >>> loader
    VanillaDatasetLoader()

    ```
    """

    @abstractmethod
    def load(self, uri: str) -> BaseDataset[T]:
        r"""Load a dataset from its Uniform Resource Identifier (URI).

        Args:
            uri: The URI of the dataset to load.

        Returns:
            The loaded dataset.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import create_vanilla_dataset
        >>> from iden.dataset.loader import VanillaDatasetLoader
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     uri = Path(tmpdir).joinpath("uri").as_uri()
        ...     _ = create_vanilla_dataset(uri=uri, shards=shards, assets=assets)
        ...     loader = VanillaDatasetLoader()
        ...     dataset = loader.load(uri)
        ...     dataset
        ...
        VanillaDataset(
          (uri): file:///.../uri
          (shards): ShardDict(
              (train): ShardTuple(
                  (0): JsonShard(uri=file:///.../shard/uri1)
                  (1): JsonShard(uri=file:///.../shard/uri2)
                )
              (val): ShardTuple()
            )
          (assets): ShardDict(
              (stats): JsonShard(uri=file:///.../uri_stats)
            )
        )

        ```
        """


def is_dataset_loader_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDatasetLoader``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseDatasetLoader`` object.

    Example usage:

    ```pycon
    >>> from iden.dataset.loader import is_dataset_loader_config
    >>> is_dataset_loader_config({"_target_": "iden.dataset.loader.VanillaDatasetLoader"})
    True

    ```
    """
    return is_object_config(config, BaseDatasetLoader)


def setup_dataset_loader(dataset_loader: BaseDatasetLoader | dict) -> BaseDatasetLoader:
    r"""Set up a dataset loader.

    The dataset loader is instantiated from its configuration by using the
    ``BaseDatasetLoader`` factory function.

    Args:
        dataset_loader: Specifies the dataset loader or its configuration.

    Returns:
        The instantiated dataset loader.

    Example usage:

    ```pycon
    >>> from iden.dataset.loader import setup_dataset_loader
    >>> dataset_loader = setup_dataset_loader({"_target_": "iden.dataset.loader.VanillaDatasetLoader"})
    >>> dataset_loader
    VanillaDatasetLoader()

    ```
    """
    if isinstance(dataset_loader, dict):
        logger.debug("Initializing a dataset loader from its configuration...")
        dataset_loader = BaseDatasetLoader.factory(**dataset_loader)
    if not isinstance(dataset_loader, BaseDatasetLoader):
        logger.warning(
            f"dataset loader is not a BaseDatasetLoader (received: {type(dataset_loader)})"
        )
    return dataset_loader



r"""Contain a shard to manage a tuple of shards."""

from __future__ import annotations

__all__ = ["ShardTuple", "create_shard_tuple"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_sequence, str_indent, str_sequence
from objectory import OBJECT_TARGET

from iden.constants import LOADER, SHARDS
from iden.io import JsonSaver, load_json
from iden.shard.base import BaseShard
from iden.shard.utils import get_list_uris
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from collections.abc import Iterable

logger = logging.getLogger(__name__)


class ShardTuple(BaseShard[tuple[BaseShard, ...]]):
    r"""Implement a data structure to manage a tuple of shards.

    Args:
        uri: The shard's URI.
        shards: The tuple of shards.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard
    >>> from iden.shard import ShardTuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = [
    ...         create_json_shard([1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()),
    ...         create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     ]
    ...     sl = ShardTuple(uri=Path(tmpdir).joinpath("main_uri").as_uri(), shards=shards)
    ...     sl
    ...
    ShardTuple(
      (0): JsonShard(uri=file:///.../shard/uri1)
      (1): JsonShard(uri=file:///.../shard/uri2)
    )

    ```
    """

    def __init__(self, uri: str, shards: Iterable[BaseShard]) -> None:
        self._uri = uri
        self._shards = tuple(shards)

    def __getitem__(self, index: int) -> BaseShard:
        return self._shards[index]

    def __len__(self) -> int:
        return len(self._shards)

    def __repr__(self) -> str:
        args = f"\n  {repr_indent(repr_sequence(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def __str__(self) -> str:
        args = f"\n  {str_indent(str_sequence(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(
            self.get_uri(), other.get_uri(), equal_nan=equal_nan
        ) and objects_are_equal(self.get_data(), other.get_data(), equal_nan=equal_nan)

    def get(self, index: int) -> BaseShard:
        r"""Get a shard.

        Args:
            index: The shard index to get.

        Returns:
            The shard.

        Raises:
            IndexError: if the index is outside  the tuple range.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard import ShardTuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     sl = ShardTuple(uri=Path(tmpdir).joinpath("main_uri").as_uri(), shards=shards)
        ...     sl.get(0)
        ...
        JsonShard(uri=file:///.../uri1)

        ```
        """
        return self[index]

    def get_data(self) -> tuple[BaseShard, ...]:
        return self._shards

    def get_uri(self) -> str:
        return self._uri

    def is_sorted_by_uri(self) -> bool:
        r"""Indicate if the shards are sorted by ascending order of URIs
        or not.

        Returns:
            ``True`` if the shards are sorted by ascending order of
                URIs, otherwise ``False``.
        """
        uris = get_list_uris(self._shards)
        return uris == sorted(uris)

    @classmethod
    def from_uri(cls, uri: str) -> ShardTuple:
        r"""Instantiate a shard from its URI.

        Args:
            uri: The URI.

        Returns:
            The instantiated shard.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import ShardTuple, create_json_shard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
        ...     create_shard_tuple(shards, uri=Path(tmpdir).joinpath("my_uri").as_uri())
        ...     shard = ShardTuple.from_uri(uri)
        ...     shard
        ...
        ShardTuple(
          (0): JsonShard(uri=file:///.../shard/uri1)
          (1): JsonShard(uri=file:///.../shard/uri2)
        )

        ```
        """
        # local import to avoid cyclic dependencies
        from iden.shard.loading import load_from_uri

        config = load_json(sanitize_path(uri))
        shards = [load_from_uri(shard) for shard in config[SHARDS]]
        return cls(uri=uri, shards=shards)

    @classmethod
    def generate_uri_config(cls, shards: Iterable[BaseShard]) -> dict:
        r"""Generate the minimal config that is used to load the shard
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            shards: The shards.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import ShardTuple, create_json_shard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     ShardTuple.generate_uri_config(shards)
        ...
        {'shards': ['file:///.../shard/uri1', 'file:///.../shard/uri2'],
         'loader': {'_target_': 'iden.shard.loader.ShardTupleLoader'}}

        ```
        """
        return {
            SHARDS: get_list_uris(shards),
            LOADER: {OBJECT_TARGET: "iden.shard.loader.ShardTupleLoader"},
        }


def create_shard_tuple(shards: Iterable[BaseShard], uri: str) -> ShardTuple:
    r"""Create a ``ShardTuple`` a list of shards.

    Note:
        It is a utility function to create a ``ShardTuple`` from its
            shards and URI. It is possible to create a ``ShardTuple``
            in other ways.

    Args:
        shards: The shards.
        uri: The shard's URI.

    Returns:
        The ``ShardTuple`` object.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import ShardTuple, create_json_shard, create_shard_tuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = [
    ...         create_json_shard(
    ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...         ),
    ...         create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     ]
    ...     shard = create_shard_tuple(shards, uri=Path(tmpdir).joinpath("my_uri").as_uri())
    ...     shard
    ...
    ShardTuple(
      (0): JsonShard(uri=file:///.../shard/uri1)
      (1): JsonShard(uri=file:///.../shard/uri2)
    )

    ```
    """
    logger.info(f"Saving URI file {uri}")
    JsonSaver().save(ShardTuple.generate_uri_config(shards), sanitize_path(uri))
    return ShardTuple(uri, shards)



r"""Contain a data structure to manage a list of shards."""

from __future__ import annotations

__all__ = ["ShardList"]

import copy
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_sequence, str_indent, str_sequence

if TYPE_CHECKING:
    from iden.shard.base import BaseShard


class ShardList:
    r"""Implement a data structure to manage a list of shards.

    Args:
        shards: The list of shards.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard
    >>> from iden.shard.collection import ShardList
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = [
    ...         create_json_shard([1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()),
    ...         create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     ]
    ...     sl = ShardList(shards)
    ...     sl.append(
    ...         create_json_shard([8, 9], uri=Path(tmpdir).joinpath("shard/uri3").as_uri())
    ...     )
    ...     sl
    ...
    ShardList(
      (0): JsonShard(uri=file:///.../shard/uri1)
      (1): JsonShard(uri=file:///.../shard/uri2)
      (2): JsonShard(uri=file:///.../shard/uri3)
    )

    ```
    """

    def __init__(self, shards: list[BaseShard] | None = None) -> None:
        self._shards = copy.copy(shards) or []

    def __getitem__(self, index: int) -> BaseShard:
        return self._shards[index]

    def __len__(self) -> int:
        return len(self._shards)

    def __repr__(self) -> str:
        args = f"\n  {repr_indent(repr_sequence(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def __str__(self) -> str:
        args = f"\n  {str_indent(str_sequence(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def append(self, shard: BaseShard) -> None:
        r"""Append a shard to the list of shards.

        Args:
            shard: The shard to add.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardList
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     sl = ShardList(shards)
        ...     sl.append(
        ...         create_json_shard([8, 9], uri=Path(tmpdir).joinpath("shard/uri3").as_uri())
        ...     )
        ...     sl
        ...
        ShardList(
          (0): JsonShard(uri=file:///.../shard/uri1)
          (1): JsonShard(uri=file:///.../shard/uri2)
          (2): JsonShard(uri=file:///.../shard/uri3)
        )

        ```
        """
        self._shards.append(shard)

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two objects are equal.

        Args:
            other: The object to compare with.
            equal_nan: If ``True``, then two ``NaN``s will be
                considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``.

        Example usage:

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardList
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     ShardList(shards).equal(ShardList(shards))
        ...
        True

        ```
        """
        if not isinstance(other, ShardList):
            return False
        return objects_are_equal(self._shards, other._shards, equal_nan=equal_nan)

    def get(self, index: int) -> BaseShard:
        r"""Get a shard.

        Args:
            index: The shard index to get.

        Returns:
            The shard.

        Raises:
            IndexError: if the index is outside  the list range.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardList
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     sl = ShardList(shards)
        ...     sl.get(0)
        ...
        JsonShard(uri=file:///.../uri1)

        ```
        """
        return self[index]

    def get_shards(self) -> list[BaseShard]:
        r"""Get the list of shards.

        Returns:
            The list of shards.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardList
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     sl = ShardList(shards)
        ...     sl.get_shards()
        ...
        [JsonShard(uri=file:///.../uri1), JsonShard(uri=file:///.../uri2)]

        ```
        """
        return self._shards

    def pop(self, index: int) -> BaseShard:
        r"""Remove the shard at the given position in the list, and
        return it.

        Args:
            index: The index of the shard to remove.

        Returns:
            The popped shard.

        Raises:
            IndexError: if the list is empty or the index is outside
                the list range.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardList
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     sl = ShardList(shards)
        ...     sl.pop(0)
        ...     sl
        ...
        JsonShard(uri=file:///.../shard/uri1)
        ShardList(
          (0): JsonShard(uri=file:///.../shard/uri2)
        )

        ```
        """
        return self._shards.pop(index)

    def get_uris(self) -> list[str]:
        r"""Get the list of shard's URI.

        Returns:
            The list of shard's URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardList
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = [
        ...         create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     ]
        ...     sl = ShardList(shards)
        ...     sl.get_uris()
        ...
        ['file:///.../shard/uri1', 'file:///.../shard/uri2']

        ```
        """
        return [shard.get_uri() for shard in self._shards]



r"""Contain code to load a shard from its Uniform Resource Identifier
(URI)."""

from __future__ import annotations

__all__ = ["load_from_uri"]

from typing import TYPE_CHECKING

from iden.constants import LOADER
from iden.io import load_json
from iden.shard.loader import setup_shard_loader
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from iden.shard import BaseShard


def load_from_uri(uri: str) -> BaseShard:
    r"""Load a shard from its Uniform Resource Identifier (URI).

    Args:
        uri: The URI of the shard.

    Returns:
        The shard associated to the URI.

    Raises:
        FileNotFoundError: if the URI file does not exist.
    """
    path = sanitize_path(uri)
    if not path.is_file():
        msg = f"uri file does not exist: {path}"
        raise FileNotFoundError(msg)
    config = load_json(path)
    loader = setup_shard_loader(config[LOADER])
    return loader.load(uri)



r"""Contain utility functions to measure time."""

from __future__ import annotations

__all__ = ["sync_perf_counter", "timeblock"]

import logging
import time
from contextlib import contextmanager
from typing import TYPE_CHECKING
from unittest.mock import Mock

from coola.utils import is_torch_available

from iden.utils.format import human_time

if is_torch_available():
    import torch
else:  # pragma: no cover
    torch = Mock()


if TYPE_CHECKING:
    from collections.abc import Generator


logger = logging.getLogger(__name__)


def sync_perf_counter() -> float:
    r"""Extension of ``time.perf_counter`` that waits for all kernels in
    all streams on a CUDA device to complete.

    Returns:
        Same as ``time.perf_counter()``.
            See https://docs.python.org/3/library/time.html#time.perf_counter
            for more information.

    Example usage:

    ```pycon

    >>> from iden.utils.time import sync_perf_counter
    >>> tic = sync_perf_counter()
    >>> x = [1, 2, 3]
    >>> toc = sync_perf_counter()
    >>> toc - tic

    ```
    """
    if is_torch_available() and torch.cuda.is_available():
        torch.cuda.synchronize()
    return time.perf_counter()


@contextmanager
def timeblock(message: str = "Total time: {time}") -> Generator[None, None, None]:
    r"""Implement a context manager to measure the execution time of a
    block of code.

    Args:
        message: The message displayed when the time is logged.

    Example usage:

    ```pycon

    >>> from iden.utils.time import timeblock
    >>> with timeblock():
    ...     x = [1, 2, 3]
    ...
    >>> with timeblock("Training: {time}"):
    ...     y = [1, 2, 3]
    ...

    ```
    """
    if "{time}" not in message:
        msg = f"{{time}} is missing in the message (received: {message})"
        raise RuntimeError(msg)
    start_time = sync_perf_counter()
    try:
        yield
    finally:
        logger.info(message.format(time=human_time(sync_perf_counter() - start_time)))



r"""Contain shard loader implementations."""

from __future__ import annotations

__all__ = ["ShardDictLoader"]


from iden.shard.base import BaseShard
from iden.shard.dict import ShardDict
from iden.shard.loader.base import BaseShardLoader


class ShardDictLoader(BaseShardLoader[dict[str, BaseShard]]):
    r"""Implement a ``ShardDict`` loader.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard, create_shard_dict
    >>> from iden.shard.loader import ShardDictLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     shards = {
    ...         "train": create_json_shard(
    ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...         ),
    ...         "val": create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     }
    ...     _ = create_shard_dict(shards, uri=Path(tmpdir).joinpath("my_uri").as_uri())
    ...     loader = ShardDictLoader()
    ...     shard = loader.load(uri)
    ...     shard
    ...
    ShardDict(
      (train): JsonShard(uri=file:///.../shard/uri1)
      (val): JsonShard(uri=file:///.../shard/uri2)
    )

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> ShardDict:
        return ShardDict.from_uri(uri)



r"""Contain the definition of the exceptions."""

from __future__ import annotations

__all__ = ["ShardExistsError", "ShardNotFoundError"]


class ShardExistsError(Exception):
    r"""Raised when trying to add a shard that already exists."""


class ShardNotFoundError(Exception):
    r"""Raised when trying to access a shard that does not exist."""



r"""Contain a data structure to manage a dictionary of shards."""

from __future__ import annotations

__all__ = ["ShardDict"]

import copy
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping

from iden.shard.exceptions import ShardExistsError, ShardNotFoundError

if TYPE_CHECKING:
    from iden.shard.base import BaseShard


class ShardDict:
    r"""Implement a data structure to manage a dictionary of shards.

    Args:
        shards: The dictionary of shards.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import VanillaDataset
    >>> from iden.shard import create_json_shard
    >>> from iden.shard.collection import ShardDict
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = {
    ...         "train": create_json_shard(
    ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...         ),
    ...         "val": create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     }
    ...     sd = ShardDict(shards)
    ...     shard = create_json_shard([8, 9], uri=Path(tmpdir).joinpath("shard/uri3").as_uri())
    ...     sd.add_shard("test", shard)
    ...     sd
    ...
    ShardDict(
      (train): JsonShard(uri=file:///.../shard/uri1)
      (val): JsonShard(uri=file:///.../shard/uri2)
      (test): JsonShard(uri=file:///.../shard/uri3)
    )

    ```
    """

    def __init__(self, shards: dict[str, BaseShard] | None = None) -> None:
        self._shards = copy.copy(shards) or {}

    def __len__(self) -> int:
        return len(self._shards)

    def __repr__(self) -> str:
        args = f"\n  {repr_indent(repr_mapping(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def __str__(self) -> str:
        args = f"\n  {str_indent(str_mapping(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def add_shard(self, shard_id: str, shard: BaseShard, replace_ok: bool = False) -> None:
        r"""Add a shard to the shard collection.

        Note that the name should be unique. If the name exists, the
        old shard will be overwritten by the new shard.

        Args:
            shard_id: The shard ID.
            shard: The shard to add to the collection.
            replace_ok: If ``False``, ``ShardExistsError`` is raised
                if a shard with the same ID exists.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(shards)
        ...     sd.add_shard("test",
        ...         create_json_shard([8, 9], uri=Path(tmpdir).joinpath("shard/uri3").as_uri())
        ...     )
        ...     sd.has_shard("test")
        ...
        True

        ```
        """
        if shard_id in self._shards and not replace_ok:
            msg = (
                f"`{shard_id}` is already used to register a shard. "
                "Use `replace_ok=True` to replace a shard"
            )
            raise ShardExistsError(msg)
        self._shards[shard_id] = shard

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two objects are equal.

        Args:
            other: The object to compare with.
            equal_nan: If ``True``, then two ``NaN``s will be
                considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``.

        Example usage:

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     ShardDict(shards).equal(ShardDict(shards))
        ...
        True

        ```
        """
        if not isinstance(other, ShardDict):
            return False
        return objects_are_equal(self._shards, other._shards, equal_nan=equal_nan)

    def get_shard(self, shard_id: str) -> Any:
        r"""Get a shard.

        Args:
            shard_id: The shard ID.

        Returns:
            The shard.

        Raises:
            ShardNotFoundError: if the shard does not exist.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(shards)
        ...     sd.get_shard("train")
        ...
        JsonShard(uri=file:///.../uri1)

        ```
        """
        if shard_id not in self._shards:
            msg = f"shard `{shard_id}` does not exist"
            raise ShardNotFoundError(msg)
        return self._shards[shard_id]

    def get_shards(self) -> dict[str, BaseShard]:
        r"""Get the dictionary of shards.

        Returns:
            The dictionary of shards.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(shards)
        ...     sd.get_shards()
        ...
        {'train': JsonShard(uri=file:///.../uri1), 'val': JsonShard(uri=file:///.../uri2)}

        ```
        """
        return self._shards

    def has_shard(self, shard_id: str) -> bool:
        r"""Indicate if the shard exists or not.

        Args:
            shard_id: The shard ID.

        Returns:
            ``True`` if the shard exists, otherwise ``False``

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(shards)
        ...     sd.has_shard("train")
        ...     sd.has_shard("test")
        ...
        True
        False

        ```
        """
        return shard_id in self._shards

    def remove_shard(self, shard_id: str) -> None:
        r"""Remove a shard.

        Args:
            shard_id: The shard ID.

        Raises:
            ShardNotFoundError: if the shard does not exist.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(shards)
        ...     sd.remove_shard("train")
        ...     sd.has_shard("train")
        ...
        False

        ```
        """
        if shard_id not in self._shards:
            msg = f"shard `{shard_id}` does not exist so it is not possible to remove it"
            raise ShardNotFoundError(msg)
        del self._shards[shard_id]

    def get_uris(self) -> dict[str, str]:
        r"""Get the dictionary of shard's URI.

        Returns:
            The dictionary of shard's URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.collection import ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(shards)
        ...     sd.get_uris()
        ...
        {'train': 'file:///.../shard/uri1', 'val': 'file:///.../shard/uri2'}

        ```
        """
        return {key: shard.get_uri() for key, shard in self._shards.items()}



r"""Contain data loaders and savers."""

from __future__ import annotations

__all__ = [
    "AutoFileLoader",
    "BaseFileSaver",
    "BaseLoader",
    "BaseSaver",
    "JsonLoader",
    "JsonSaver",
    "PickleLoader",
    "PickleSaver",
    "TextLoader",
    "TextSaver",
    "TorchLoader",
    "TorchSaver",
    "YamlLoader",
    "YamlSaver",
    "is_loader_config",
    "is_saver_config",
    "load_json",
    "load_pickle",
    "load_text",
    "load_torch",
    "load_yaml",
    "save_json",
    "save_pickle",
    "save_text",
    "save_torch",
    "save_yaml",
    "setup_loader",
    "setup_saver",
]

from iden.io.auto import AutoFileLoader, register_auto_loaders
from iden.io.base import (
    BaseFileSaver,
    BaseLoader,
    BaseSaver,
    is_loader_config,
    is_saver_config,
    setup_loader,
    setup_saver,
)
from iden.io.json import JsonLoader, JsonSaver, load_json, save_json
from iden.io.pickle import PickleLoader, PickleSaver, load_pickle, save_pickle
from iden.io.text import TextLoader, TextSaver, load_text, save_text
from iden.io.torch import TorchLoader, TorchSaver, load_torch, save_torch
from iden.io.yaml import YamlLoader, YamlSaver, load_yaml, save_yaml

register_auto_loaders()



r"""Contain shard loader implementations."""

from __future__ import annotations

__all__ = [
    "BaseShardLoader",
    "FileShardLoader",
    "JsonShardLoader",
    "PickleShardLoader",
    "ShardDictLoader",
    "ShardTupleLoader",
    "TorchSafetensorsShardLoader",
    "TorchShardLoader",
    "YamlShardLoader",
    "is_shard_loader_config",
    "setup_shard_loader",
]

from iden.shard.loader.base import (
    BaseShardLoader,
    is_shard_loader_config,
    setup_shard_loader,
)
from iden.shard.loader.dict import ShardDictLoader
from iden.shard.loader.file import FileShardLoader
from iden.shard.loader.json import JsonShardLoader
from iden.shard.loader.pickle import PickleShardLoader
from iden.shard.loader.safetensors import TorchSafetensorsShardLoader
from iden.shard.loader.torch import TorchShardLoader
from iden.shard.loader.tuple import ShardTupleLoader
from iden.shard.loader.yaml import YamlShardLoader



r"""Contain the base class to implement a dataset object."""

from __future__ import annotations

__all__ = ["BaseDataset"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:

    from iden.shard import BaseShard

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BaseDataset(Generic[T], ABC):
    r"""Define the base class to implement a dataset.

    Note this dataset class is very different from the PyTorch dataset
    class because it has a different goal. One of the goals is to help
    to organize and manage shards.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import VanillaDataset
    >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = create_shard_dict(
    ...         shards={
    ...             "train": create_shard_tuple(
    ...                 [
    ...                     create_json_shard(
    ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...                     ),
    ...                     create_json_shard(
    ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...                     ),
    ...                 ],
    ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
    ...             ),
    ...             "val": create_shard_tuple(
    ...                 shards=[],
    ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
    ...             ),
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
    ...     )
    ...     assets = create_shard_dict(
    ...         shards={
    ...             "stats": create_json_shard(
    ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
    ...             )
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
    ...     )
    ...     dataset = VanillaDataset(
    ...         uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets
    ...     )
    ...     dataset
    ...
    VanillaDataset(
      (uri): file:///.../uri
      (shards): ShardDict(
          (train): ShardTuple(
              (0): JsonShard(uri=file:///.../shard/uri1)
              (1): JsonShard(uri=file:///.../shard/uri2)
            )
          (val): ShardTuple()
        )
      (assets): ShardDict(
          (stats): JsonShard(uri=file:///.../uri_stats)
        )
    )

    ```
    """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two datasets are equal or not.

        Args:
            other: The object to compare with.
            equal_nan: If ``True``, then two ``NaN``s will be
                considered equal.

        Returns:
            ``True`` if the two datasets are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset1 = VanillaDataset(
        ...         uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets
        ...     )
        ...     dataset2 = VanillaDataset(
        ...         uri=Path(tmpdir).joinpath("uri2").as_uri(), shards=shards, assets=assets
        ...     )
        ...     dataset1.equal(dataset2)
        ...
        False

        ```
        """

    @abstractmethod
    def get_asset(self, asset_id: str) -> Any:
        r"""Get a data asset from this sharded dataset.

        This method is useful to access some data variables/parameters
        that are not available before to load/preprocess the data.

        Args:
            asset_id: The asset ID used to find the asset.

        Returns:
            The asset.

        Raises:
            AssetNotFoundError: if the asset does not exist.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     dataset.get_asset('stats').get_data()
        ...
        {'mean': 42}

        ```
        """

    @abstractmethod
    def has_asset(self, asset_id: str) -> bool:
        r"""Indicate if the asset exists or not.

        Args:
            asset_id: The asset ID used to find the asset.

        Returns:
            ``True`` if the asset exists, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     dataset.has_asset('stats')
        ...     dataset.has_asset('missing')
        ...
        True
        False

        ```
        """

    @abstractmethod
    def get_shards(self, split: str) -> tuple[BaseShard[T], ...]:
        r"""Get the shards for a given split.

        Returns:
            The shards for a given split. The shards are
                sorted by ascending order of URI.

        Raises:
            ``SplitNotFoundError``: if the split does not exist.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     dataset.get_shards('train')
        ...     dataset.get_shards('val')
        ...
        (JsonShard(uri=file:///.../uri1), JsonShard(uri=file:///.../uri2))
        ()

        ```
        """

    @abstractmethod
    def get_num_shards(self, split: str) -> int:
        r"""Get the number of shards for a given split.

        Returns:
            The number of shards in the dataset for a given split.

        Raises:
            ``SplitNotFoundError``: if the split does not exist.

        Returns:
            The dataset splits.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     dataset.get_num_shards('train')
        ...     dataset.get_num_shards('val')
        ...
        2
        0

        ```
        """

    @abstractmethod
    def get_splits(self) -> set[str]:
        r"""Get the available dataset splits.

        Returns:
            The dataset splits.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     sorted(dataset.get_splits())
        ...
        ['train', 'val']

        ```
        """

    @abstractmethod
    def has_split(self, split: str) -> bool:
        r"""Indicate if a dataset split exists or not.

        Returns:
            ``True`` of the split exists, otherwise ``False``

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     dataset.has_split('train')
        ...     dataset.has_split('missing')
        ...
        True
        False

        ```
        """

    @abstractmethod
    def get_uri(self) -> str:
        r"""Get the Uniform Resource Identifier (URI) of the dataset.

        Returns:
            The dataset's URI.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 {'mean': 42}, uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_asset").as_uri(),
        ...     )
        ...     dataset = VanillaDataset(uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets)
        ...     dataset.get_uri()
        ...
        file:///.../uri

        ```
        """



r"""Contain pickle-based shard implementations."""

from __future__ import annotations

__all__ = ["PickleShard", "create_pickle_shard"]

import logging
from typing import TYPE_CHECKING, Any

from objectory import OBJECT_TARGET

from iden.constants import KWARGS, LOADER
from iden.io import JsonSaver, PickleLoader, PickleSaver
from iden.shard.file import FileShard
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class PickleShard(FileShard[Any]):
    r"""Implement a pickle shard.

    The data are stored in a pickle file.

    Args:
        uri: The shard's URI.
        path: Specifies the path to the pickle file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import PickleShard
    >>> from iden.io import save_pickle
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     file = Path(tmpdir).joinpath("data.pkl")
    ...     save_pickle([1, 2, 3], file)
    ...     shard = PickleShard(uri="file:///data/1234456789", path=file)
    ...     shard.get_data()
    ...
    [1, 2, 3]

    ```
    """

    def __init__(self, uri: str, path: Path | str) -> None:
        super().__init__(uri, path, loader=PickleLoader())

    @classmethod
    def generate_uri_config(cls, path: Path) -> dict:
        r"""Generate the minimal config that is used to load the shard
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            path: The path to the pickle file.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import PickleShard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     file = Path(tmpdir).joinpath("data.pkl")
        ...     PickleShard.generate_uri_config(file)
        ...
        {'kwargs': {'path': '.../data.pkl'},
         'loader': {'_target_': 'iden.shard.loader.PickleShardLoader'}}

        ```
        """
        return {
            KWARGS: {"path": sanitize_path(path).as_posix()},
            LOADER: {OBJECT_TARGET: "iden.shard.loader.PickleShardLoader"},
        }


def create_pickle_shard(data: Any, uri: str, path: Path | None = None) -> PickleShard:
    r"""Create a ``PickleShard`` from data.

    Note:
        It is a utility function to create a ``PickleShard`` from its
            data and URI. It is possible to create a ``PickleShard``
            in other ways.

    Args:
        data: The data to save in the pickle file.
        uri: The shard's URI.
        path: The path to the pickle file. If ``None``, a path is
            automatically based on the URI.

    Returns:
        The ``PickleShard`` object.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_pickle_shard
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shard = create_pickle_shard([1, 2, 3], uri=Path(tmpdir).joinpath("my_uri").as_uri())
    ...     shard.get_data()
    ...
    [1, 2, 3]

    ```
    """
    if path is None:
        path = sanitize_path(uri + ".pkl")
    logger.info(f"Saving URI file {uri}")
    JsonSaver().save(PickleShard.generate_uri_config(path), sanitize_path(uri))
    logger.info(f"Saving data in file {path}")
    PickleSaver().save(data, path)
    return PickleShard(uri, path)



r"""Contain YAML-based data loaders and savers."""

from __future__ import annotations

__all__ = ["YamlLoader", "YamlSaver", "load_yaml", "save_yaml", "get_loader_mapping"]

from pathlib import Path
from typing import Any, TypeVar
from unittest.mock import Mock

from iden.io.base import BaseFileSaver, BaseLoader
from iden.utils.imports import check_yaml, is_yaml_available

if is_yaml_available():
    import yaml
else:  # pragma: no cover
    yaml = Mock()


T = TypeVar("T")


class YamlLoader(BaseLoader[Any]):
    r"""Implement a data loader to load data in a YAML file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import load_yaml
    >>> data = YamlLoader().load(Path("/path/to/data.yaml"))  # xdoctest: +SKIP()

    ```
    """

    def __init__(self) -> None:
        check_yaml()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, path: Path) -> Any:
        with Path.open(path, mode="rb") as file:
            return yaml.safe_load(file)


class YamlSaver(BaseFileSaver[Any]):
    r"""Implement a file saver to save data with a YAML file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import YamlSaver
    >>> YamlSaver().save({"key": "value"}, Path("/path/to/data.yaml"))  # xdoctest: +SKIP()

    ```
    """

    def __init__(self) -> None:
        check_yaml()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _save_file(self, to_save: Any, path: Path) -> None:
        # Save to tmp, then commit by moving the file in case the job gets
        # interrupted while writing the file
        tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
        with Path.open(tmp_path, mode="w") as file:
            yaml.dump(to_save, file, Dumper=yaml.Dumper)
        tmp_path.rename(path)


def load_yaml(path: Path) -> Any:
    r"""Load the data from a given YAML file.

    Args:
        path: Specifies the path to the YAML file.

    Returns:
        The data from the YAML file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import load_yaml
    >>> data = load_yaml(Path("/path/to/data.yaml"))  # xdoctest: +SKIP()

    ```
    """
    return YamlLoader().load(path)


def save_yaml(to_save: Any, path: Path, *, exist_ok: bool = False) -> None:
    r"""Save the given data in a YAML file.

    Args:
        to_save: Specifies the data to write in a YAML file.
        path: Specifies the path where to write the YAML file.
        exist_ok: If ``exist_ok`` is ``False`` (the default),
            ``FileExistsError`` is raised if the target file
            already exists. If ``exist_ok`` is ``True``,
            ``FileExistsError`` will not be raised unless the
            given path already exists in the file system and is
            not a file.

    Raises:
        FileExistsError: if the file already exists.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import save_yaml
    >>> save_yaml({"key": "value"}, Path("/path/to/data.yaml"))  # xdoctest: +SKIP()

    ```
    """
    YamlSaver().save(to_save, path, exist_ok=exist_ok)


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.yaml import get_loader_mapping
    >>> get_loader_mapping()
    {'yaml': YamlLoader(), 'yml': YamlLoader()}

    ```
    """
    if not is_yaml_available():
        return {}
    loader = YamlLoader()
    return {"yaml": loader, "yml": loader}



r"""Contain safetensors data loaders and savers."""

from __future__ import annotations

__all__ = [
    "NumpyLoader",
    "NumpySafetensorsLoader",
    "NumpySafetensorsSaver",
    "NumpySaver",
    "TorchLoader",
    "TorchSafetensorsLoader",
    "TorchSafetensorsSaver",
    "TorchSaver",
]

from iden.io.safetensors.loaders import NumpySafetensorsLoader
from iden.io.safetensors.loaders import NumpySafetensorsLoader as NumpyLoader
from iden.io.safetensors.loaders import TorchSafetensorsLoader
from iden.io.safetensors.loaders import TorchSafetensorsLoader as TorchLoader
from iden.io.safetensors.savers import NumpySafetensorsSaver
from iden.io.safetensors.savers import NumpySafetensorsSaver as NumpySaver
from iden.io.safetensors.savers import TorchSafetensorsSaver
from iden.io.safetensors.savers import TorchSafetensorsSaver as TorchSaver



r"""Contain path utility functions."""

from __future__ import annotations

__all__ = ["sanitize_path"]

from pathlib import Path
from urllib.parse import unquote, urlparse


def sanitize_path(path: Path | str) -> Path:
    r"""Sanitize the given path.

    Args:
        path: Specifies the path to sanitize.

    Returns:
        The sanitized path.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.utils.path import sanitize_path
    >>> sanitize_path("something")
    PosixPath('.../something')
    >>> sanitize_path("")
    PosixPath('...')
    >>> sanitize_path(Path("something"))
    PosixPath('.../something')
    >>> sanitize_path(Path("something/./../"))
    PosixPath('...')

    ```
    """
    if isinstance(path, str):
        # Use urlparse to parse file URI: https://stackoverflow.com/a/15048213
        path = Path(unquote(urlparse(path).path))
    return path.expanduser().resolve()



r"""Contain shard loader implementations."""

from __future__ import annotations

__all__ = ["ShardTupleLoader"]


from iden.shard.base import BaseShard
from iden.shard.loader.base import BaseShardLoader
from iden.shard.tuple import ShardTuple


class ShardTupleLoader(BaseShardLoader[tuple[BaseShard, ...]]):
    r"""Implement a ``ShardTuple`` loader.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard, create_shard_tuple
    >>> from iden.shard.loader import ShardTupleLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     shards = [
    ...         create_json_shard([1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()),
    ...         create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     ]
    ...     _ = create_shard_tuple(shards, uri=Path(tmpdir).joinpath("my_uri").as_uri())
    ...     loader = ShardTupleLoader()
    ...     shard = loader.load(uri)
    ...     shard
    ...
    ShardTuple(
      (0): JsonShard(uri=file:///.../shard/uri1)
      (1): JsonShard(uri=file:///.../shard/uri2)
    )

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> ShardTuple:
        return ShardTuple.from_uri(uri)



r"""Define some PyTest fixtures."""

from __future__ import annotations

__all__ = ["safetensors_available", "yaml_available"]

import pytest

from iden.utils.imports import is_safetensors_available, is_yaml_available

safetensors_available = pytest.mark.skipif(
    not is_safetensors_available(), reason="Require safetensors"
)
yaml_available = pytest.mark.skipif(not is_yaml_available(), reason="Require yaml")



r"""Contain the base class to implement a data loader or saver
object."""

from __future__ import annotations

__all__ = [
    "BaseLoader",
    "BaseSaver",
    "BaseFileSaver",
    "is_loader_config",
    "is_saver_config",
    "setup_loader",
    "setup_saver",
]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    from pathlib import Path

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseLoader(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a data loader."""

    @abstractmethod
    def load(self, path: Path) -> T:
        r"""Save the data into the given path.

        Args:
            path: The path with the data to load.

        Returns:
            The data
        """


class BaseSaver(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a data saver."""

    @abstractmethod
    def save(self, to_save: T, path: Path, exist_ok: bool = False) -> None:
        r"""Save the data into the given path.

        Args:
            to_save: The data to save. The data should be compatible
                with the saving engine.
            path: Specifies the path where to save the data.
            exist_ok: If ``exist_ok`` is ``False`` (the default),
                an exception is raised if the target path already
                exists.
        """


class BaseFileSaver(BaseSaver[T]):
    r"""Define the base class to implement a file saver."""

    def save(self, to_save: T, path: Path, exist_ok: bool = False) -> None:
        r"""Save the data into the given path.

        Args:
            to_save: The data to save. The data should be compatible
                with the saving engine.
            path: Specifies the path where to save the data.
            exist_ok: If ``exist_ok`` is ``False`` (the default),
                ``FileExistsError`` is raised if the target file
                already exists. If ``exist_ok`` is ``True``,
                ``FileExistsError`` will not be raised unless the
                given path already exists in the file system and is
                not a file.

        Raises:
            FileExistsError: if the file already exists.
        """
        if path.is_dir():
            msg = f"path ({path}) is a directory"
            raise IsADirectoryError(msg)
        if path.is_file() and not exist_ok:
            msg = (
                f"path ({path}) already exists. "
                f"Please use `exist_ok=True` if you want to overwrite the setter for this name"
            )
            raise FileExistsError(msg)
        path.parent.mkdir(exist_ok=True, parents=True)
        self._save_file(to_save, path)

    @abstractmethod
    def _save_file(self, to_save: T, path: Path) -> None:
        r"""Save the data into the given file.

        Args:
            to_save: The data to save. The data should be compatible
                with the saving engine.
            path: Specifies the path where to save the data.
        """


def is_loader_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseLoader``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseLoader`` object.

    Example usage:

    ```pycon
    >>> from iden.io import is_loader_config
    >>> is_loader_config({"_target_": "iden.io.JsonLoader"})
    True

    ```
    """
    return is_object_config(config, BaseLoader)


def is_saver_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseSaver``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseSaver`` object.

    Example usage:

    ```pycon
    >>> from iden.io import is_saver_config
    >>> is_saver_config({"_target_": "iden.io.JsonSaver"})
    True

    ```
    """
    return is_object_config(config, BaseSaver)


def setup_loader(loader: BaseLoader | dict) -> BaseLoader:
    r"""Set up a data loader.

    The data loader is instantiated from its configuration by using the
    ``BaseLoader`` factory function.

    Args:
        loader: Specifies the data loader or its configuration.

    Returns:
        The instantiated data loader.

    Example usage:

    ```pycon
    >>> from iden.io import setup_loader
    >>> loader = setup_loader({"_target_": "iden.io.JsonLoader"})
    >>> loader
    JsonLoader()

    ```
    """
    if isinstance(loader, dict):
        logger.debug("Initializing a data loader from its configuration...")
        loader = BaseLoader.factory(**loader)
    if not isinstance(loader, BaseLoader):
        logger.warning(f"data loader is not a BaseLoader (received: {type(loader)})")
    return loader


def setup_saver(saver: BaseSaver | dict) -> BaseSaver:
    r"""Set up a data saver.

    The data saver is instantiated from its configuration by using the
    ``BaseSaver`` factory function.

    Args:
        saver: Specifies the data saver or its configuration.

    Returns:
        The instantiated data saver.

    Example usage:

    ```pycon
    >>> from iden.io import setup_saver
    >>> saver = setup_saver({"_target_": "iden.io.JsonSaver"})
    >>> saver
    JsonSaver()

    ```
    """
    if isinstance(saver, dict):
        logger.debug("Initializing a data saver from its configuration...")
        saver = BaseSaver.factory(**saver)
    if not isinstance(saver, BaseSaver):
        logger.warning(f"data saver is not a BaseSaver (received: {type(saver)})")
    return saver



r"""Contain code to load a shard from its Uniform Resource Identifier
(URI)."""

from __future__ import annotations

__all__ = ["get_dict_uris", "get_list_uris", "sort_by_uri"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable

    from iden.shard.base import BaseShard


def get_dict_uris(shards: dict[str, BaseShard]) -> dict[str, str]:
    r"""Get the dictionary of shard's URI.

    Args:
        shards: The dictionary of shards.

    Returns:
        The dictionary of shard's URI.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard, get_dict_uris
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = {
    ...         "train": create_json_shard(
    ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...         ),
    ...         "val": create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     }
    ...     get_dict_uris(shards)
    ...
    {'train': 'file:///.../shard/uri1', 'val': 'file:///.../shard/uri2'}

    ```
    """
    return {key: shard.get_uri() for key, shard in shards.items()}


def get_list_uris(shards: Iterable[BaseShard]) -> list[str]:
    r"""Get the list of shard's URI.

    Args:
        shards: The shards.

    Returns:
        The tuple of shard's URI.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import get_list_uris, create_json_shard
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = [
    ...         create_json_shard([1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()),
    ...         create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     ]
    ...     get_list_uris(shards)
    ...
    ['file:///.../shard/uri1', 'file:///.../shard/uri2']

    ```
    """
    return [shard.get_uri() for shard in shards]


def sort_by_uri(shards: Iterable[BaseShard], /, *, reverse: bool = False) -> list[BaseShard]:
    r"""Sort a sequence of shards by their URIs.

    Args:
        shards: The shards to sort.
        reverse: If set to ``True``, then the list elements are sorted
            as if each comparison were reversed.

    Returns:
        The sorted shards.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard, sort_by_uri
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = sort_by_uri(
    ...         [
    ...             create_json_shard([1, 2, 3], uri=Path(tmpdir).joinpath("uri2").as_uri()),
    ...             create_json_shard([4, 5, 6, 7], uri=Path(tmpdir).joinpath("uri3").as_uri()),
    ...             create_json_shard([4, 5, 6, 7], uri=Path(tmpdir).joinpath("uri1").as_uri()),
    ...         ]
    ...     )
    ...     shards
    ...
    [JsonShard(uri=file:///.../uri1), JsonShard(uri=file:///.../uri2), JsonShard(uri=file:///.../uri3)]

    ```
    """
    return sorted(shards, key=lambda item: item.get_uri(), reverse=reverse)



r"""Contain I/O utility functions."""

from __future__ import annotations

__all__ = ["get_loader_mapping"]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from iden.io import BaseLoader


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.utils import get_loader_mapping
    >>> get_loader_mapping()
    {'json': JsonLoader(), 'pkl': PickleLoader(), 'pickle': PickleLoader(), ...}

    ```
    """
    from iden import io  # Local import to avoid cyclic dependencies

    return (
        io.json.get_loader_mapping()
        | io.pickle.get_loader_mapping()
        | io.text.get_loader_mapping()
        | io.torch.get_loader_mapping()
        | io.yaml.get_loader_mapping()
    )



r"""Contain dataset loader implementations."""

from __future__ import annotations

__all__ = [
    "BaseDatasetLoader",
    "VanillaDatasetLoader",
    "is_dataset_loader_config",
    "setup_dataset_loader",
]

from iden.dataset.loader.base import (
    BaseDatasetLoader,
    is_dataset_loader_config,
    setup_dataset_loader,
)
from iden.dataset.loader.vanilla import VanillaDatasetLoader



r"""Contain shard implementations."""

from __future__ import annotations

__all__ = [
    "BaseShard",
    "FileShard",
    "InMemoryShard",
    "JsonShard",
    "PickleShard",
    "ShardDict",
    "ShardTuple",
    "TorchSafetensorsShard",
    "TorchShard",
    "YamlShard",
    "create_json_shard",
    "create_pickle_shard",
    "create_shard_dict",
    "create_shard_tuple",
    "create_torch_safetensors_shard",
    "create_torch_shard",
    "create_yaml_shard",
    "get_dict_uris",
    "get_list_uris",
    "load_from_uri",
    "sort_by_uri",
]

from iden.shard import comparators  # noqa: F401
from iden.shard.base import BaseShard
from iden.shard.dict import ShardDict, create_shard_dict
from iden.shard.file import FileShard
from iden.shard.in_memory import InMemoryShard
from iden.shard.json import JsonShard, create_json_shard
from iden.shard.loading import load_from_uri
from iden.shard.pickle import PickleShard, create_pickle_shard
from iden.shard.safetensors import TorchSafetensorsShard, create_torch_safetensors_shard
from iden.shard.torch import TorchShard, create_torch_shard
from iden.shard.tuple import ShardTuple, create_shard_tuple
from iden.shard.utils import get_dict_uris, get_list_uris, sort_by_uri
from iden.shard.yaml import YamlShard, create_yaml_shard



r"""Contain safetensors shard loader implementations."""

from __future__ import annotations

__all__ = ["TorchSafetensorsShardLoader"]

from unittest.mock import Mock

from coola.utils import check_torch, is_torch_available

from iden.shard.loader.base import BaseShardLoader
from iden.shard.safetensors import TorchSafetensorsShard
from iden.utils.imports import check_safetensors

if is_torch_available():
    import torch
else:  # pragma: no cover
    torch = Mock()


class TorchSafetensorsShardLoader(BaseShardLoader[dict[str, torch.Tensor]]):
    r"""Implement a safetensors shard loader for ``torch.Tensor``s.

    Raises:
        RuntimeError: if ``safetensors`` or ``torch`` is not installed.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> import torch
    >>> from pathlib import Path
    >>> from iden.shard import create_torch_safetensors_shard
    >>> from iden.shard.loader import TorchSafetensorsShardLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     _ = create_torch_safetensors_shard(
    ...         {"key1": torch.ones(2, 3), "key2": torch.arange(5)}, uri=uri
    ...     )
    ...     loader = TorchSafetensorsShardLoader()
    ...     shard = loader.load(uri)
    ...     shard
    ...
    TorchSafetensorsShard(uri=file:///.../my_uri)

    ```
    """

    def __init__(self) -> None:
        check_safetensors()
        check_torch()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> TorchSafetensorsShard:
        return TorchSafetensorsShard.from_uri(uri)



r"""Contain utility functions to compute formatted strings."""

from __future__ import annotations

__all__ = ["human_time"]

import datetime


def human_time(seconds: float) -> str:
    r"""Return a number of seconds in an easier format to read hh:mm:ss.

    If the number of seconds is bigger than 1 day, this representation
    also encodes the number of days.

    Args:
        seconds: The number of seconds.

    Returns:
        The number of seconds in a string format (hh:mm:ss).

    Example usage:

    ```pycon
    >>> from iden.utils.format import human_time
    >>> human_time(1.2)
    '0:00:01.200000'
    >>> human_time(61.2)
    '0:01:01.200000'
    >>> human_time(3661.2)
    '1:01:01.200000'

    ```
    """
    return str(datetime.timedelta(seconds=seconds))



r"""Contain JSON-based data loaders and savers."""

from __future__ import annotations

__all__ = ["JsonLoader", "JsonSaver", "load_json", "save_json", "get_loader_mapping"]

import json
from pathlib import Path
from typing import Any, TypeVar

from iden.io.base import BaseFileSaver, BaseLoader

T = TypeVar("T")


class JsonLoader(BaseLoader[Any]):
    r"""Implement a data loader to load data in a JSON file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import JsonLoader
    >>> data = JsonLoader().load(Path("/path/to/data.json"))  # xdoctest: +SKIP()

    ```
    """

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, path: Path) -> Any:
        with Path.open(path, mode="rb") as file:
            return json.load(file)


class JsonSaver(BaseFileSaver[Any]):
    r"""Implement a file saver to save data with a JSON file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import JsonSaver
    >>> JsonSaver().save({"key": "value"}, Path("/path/to/data.json"))  # xdoctest: +SKIP()

    ```
    """

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _save_file(self, to_save: Any, path: Path) -> None:
        # Save to tmp, then commit by moving the file in case the job gets
        # interrupted while writing the file
        tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
        with Path.open(tmp_path, "w") as file:
            json.dump(to_save, file, sort_keys=False)
        tmp_path.rename(path)


def load_json(path: Path) -> Any:
    r"""Load the data from a given JSON file.

    Args:
        path: Specifies the path to the JSON file.

    Returns:
        The data from the JSON file.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import load_json
    >>> data = load_json(Path("/path/to/data.json"))  # xdoctest: +SKIP()

    ```
    """
    return JsonLoader().load(path)


def save_json(to_save: Any, path: Path, *, exist_ok: bool = False) -> None:
    r"""Save the given data in a JSON file.

    Args:
        to_save: Specifies the data to write in a JSON file.
        path: Specifies the path where to write the JSON file.
        exist_ok: If ``exist_ok`` is ``False`` (the default),
            ``FileExistsError`` is raised if the target file
            already exists. If ``exist_ok`` is ``True``,
            ``FileExistsError`` will not be raised unless the
            given path already exists in the file system and is
            not a file.

    Raises:
        FileExistsError: if the file already exists.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import save_json
    >>> save_json({"key": "value"}, Path("/path/to/data.json"))  # xdoctest: +SKIP()

    ```
    """
    JsonSaver().save(to_save, path, exist_ok=exist_ok)


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.json import get_loader_mapping
    >>> get_loader_mapping()
    {'json': JsonLoader()}

    ```
    """
    return {"json": JsonLoader()}



r"""Define some utility functions for testing."""

from __future__ import annotations

__all__ = ["safetensors_available", "yaml_available"]

from iden.testing.fixtures import safetensors_available, yaml_available



r"""Contain shard loader implementations for
``VanillaDatasetLoader``."""

from __future__ import annotations

__all__ = ["VanillaDatasetLoader"]

from typing import TypeVar

from iden.dataset.loader.base import BaseDatasetLoader
from iden.dataset.vanilla import VanillaDataset

T = TypeVar("T")


class VanillaDatasetLoader(BaseDatasetLoader[T]):
    r"""Implement a ``VanillaDatasetLoader`` loader."""

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> VanillaDataset[T]:
        return VanillaDataset.from_uri(uri)



r"""Contain dataset implementations."""

from __future__ import annotations

__all__ = ["BaseDataset", "VanillaDataset", "create_vanilla_dataset", "load_from_uri"]

from iden.dataset.base import BaseDataset
from iden.dataset.loading import load_from_uri
from iden.dataset.vanilla import VanillaDataset, create_vanilla_dataset



r"""Contain loaders to load data in a safetensors format."""

from __future__ import annotations

__all__ = ["NumpySafetensorsLoader", "TorchSafetensorsLoader"]

from typing import TYPE_CHECKING
from unittest.mock import Mock

from coola.utils import check_numpy, check_torch, is_numpy_available, is_torch_available

from iden.io.base import BaseLoader
from iden.utils.imports import check_safetensors, is_safetensors_available
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path


if is_safetensors_available() and is_numpy_available():
    import numpy as np
    from safetensors import numpy as sn
else:  # pragma: no cover
    np = Mock()
    sn = Mock()

if is_safetensors_available() and is_torch_available():
    import torch
    from safetensors import torch as st
else:  # pragma: no cover
    st = Mock()
    torch = Mock()


class NumpySafetensorsLoader(BaseLoader[dict[str, np.ndarray]]):
    r"""Implement a file loader to load ``numpy.ndarray``s in the
    safetensors format.

    Link: https://huggingface.co/docs/safetensors/en/index
    """

    def __init__(self) -> None:
        check_safetensors()
        check_numpy()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, path: Path) -> dict[str, np.ndarray]:
        return sn.load_file(sanitize_path(path))


class TorchSafetensorsLoader(BaseLoader[dict[str, torch.Tensor]]):
    r"""Implement a file loader to load ``torch.Tensor``s in the
    safetensors format.

    Link: https://huggingface.co/docs/safetensors/en/index
    """

    def __init__(self, device: str | dict = "cpu") -> None:
        check_safetensors()
        check_torch()
        self._device = device

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(device={self._device})"

    def load(self, path: Path) -> dict[str, torch.Tensor]:
        return st.load_file(sanitize_path(path), device=self._device)



r"""Contain a data structure to manage a dictionary of shards."""

from __future__ import annotations

__all__ = ["ShardDict"]

import logging
from typing import Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping
from objectory import OBJECT_TARGET

from iden.constants import LOADER, SHARDS
from iden.io import JsonSaver, load_json
from iden.shard.base import BaseShard
from iden.shard.exceptions import ShardNotFoundError
from iden.shard.utils import get_dict_uris
from iden.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class ShardDict(BaseShard):
    r"""Implement a data structure to manage a dictionary of shards.

    Args:
        uri: The shard's URI.
        shards: The dictionary of shards.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import VanillaDataset
    >>> from iden.shard import create_json_shard, ShardDict
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = {
    ...         "train": create_json_shard(
    ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...         ),
    ...         "val": create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     }
    ...     sd = ShardDict(uri=Path(tmpdir).joinpath("main_uri").as_uri(), shards=shards)
    ...     sd
    ...
    ShardDict(
      (train): JsonShard(uri=file:///.../shard/uri1)
      (val): JsonShard(uri=file:///.../shard/uri2)
    )

    ```
    """

    def __init__(self, uri: str, shards: dict[str, BaseShard]) -> None:
        self._uri = uri
        self._shards = shards.copy()

    def __contains__(self, item: str) -> bool:
        return item in self._shards

    def __getitem__(self, item: str) -> BaseShard:
        return self._shards[item]

    def __len__(self) -> int:
        return len(self._shards)

    def __repr__(self) -> str:
        args = f"\n  {repr_indent(repr_mapping(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def __str__(self) -> str:
        args = f"\n  {str_indent(str_mapping(self._shards))}\n" if self._shards else ""
        return f"{self.__class__.__qualname__}({args})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(
            self.get_uri(), other.get_uri(), equal_nan=equal_nan
        ) and objects_are_equal(self.get_data(), other.get_data(), equal_nan=equal_nan)

    def get_data(self) -> dict[str, BaseShard]:
        return self._shards.copy()

    def get_uri(self) -> str:
        return self._uri

    def get_shard(self, shard_id: str) -> Any:
        r"""Get a shard.

        Args:
            shard_id: The shard ID.

        Returns:
            The shard.

        Raises:
            ShardNotFoundError: if the shard does not exist.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard, ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(uri=Path(tmpdir).joinpath("main_uri").as_uri(), shards=shards)
        ...     sd.get_shard("train")
        ...
        JsonShard(uri=file:///.../uri1)

        ```
        """
        shard = self._shards.get(shard_id, None)
        if shard is None:
            msg = f"shard `{shard_id}` does not exist"
            raise ShardNotFoundError(msg)
        return shard

    def get_shard_ids(self) -> set[str]:
        r"""Get the shard IDs.

        Returns:
            The shard IDs.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard, ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(uri=Path(tmpdir).joinpath("main_uri").as_uri(), shards=shards)
        ...     sorted(sd.get_shard_ids())
        ...
        ['train', 'val']

        ```
        """
        return set(self._shards.keys())

    def has_shard(self, shard_id: str) -> bool:
        r"""Indicate if the shard exists or not.

        Args:
            shard_id: The shard ID.

        Returns:
            ``True`` if the shard exists, otherwise ``False``

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard, ShardDict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     sd = ShardDict(uri=Path(tmpdir).joinpath("main_uri").as_uri(), shards=shards)
        ...     sd.has_shard("train")
        ...     sd.has_shard("test")
        ...
        True
        False

        ```
        """
        return shard_id in self

    @classmethod
    def from_uri(cls, uri: str) -> ShardDict:
        r"""Instantiate a shard from its URI.

        Args:
            uri: The URI.

        Returns:
            The instantiated shard.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import ShardDict, create_json_shard, create_shard_dict
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
        ...     _ = create_shard_dict(shards, uri=Path(tmpdir).joinpath("my_uri").as_uri())
        ...     shard = ShardDict.from_uri(uri)
        ...     shard
        ...
        ShardDict(
          (train): JsonShard(uri=file:///.../shard/uri1)
          (val): JsonShard(uri=file:///.../shard/uri2)
        )

        ```
        """
        # local import to avoid cyclic dependencies
        from iden.shard import load_from_uri

        config = load_json(sanitize_path(uri))
        shards = {key: load_from_uri(shard) for key, shard in config[SHARDS].items()}
        return cls(uri=uri, shards=shards)

    @classmethod
    def generate_uri_config(cls, shards: dict[str, BaseShard]) -> dict:
        r"""Generate the minimal config that is used to load the shard
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            shards: The shards.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import ShardDict, create_json_shard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = {
        ...         "train": create_json_shard(
        ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...         ),
        ...         "val": create_json_shard(
        ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...         ),
        ...     }
        ...     ShardDict.generate_uri_config(shards)
        ...
        {'shards': {'train': 'file:///.../shard/uri1', 'val': 'file:///.../shard/uri2'},
         'loader': {'_target_': 'iden.shard.loader.ShardDictLoader'}}

        ```
        """
        return {
            SHARDS: get_dict_uris(shards),
            LOADER: {OBJECT_TARGET: "iden.shard.loader.ShardDictLoader"},
        }


def create_shard_dict(shards: dict[str, BaseShard], uri: str) -> ShardDict:
    r"""Create a ``ShardDict`` a list of shards.

    Note:
        It is a utility function to create a ``ShardDict`` from its
            shards and URI. It is possible to create a ``ShardDict``
            in other ways.

    Args:
        shards: The shards.
        uri: The shard's URI.

    Returns:
        The ``ShardDict`` object.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import ShardDict, create_json_shard, create_shard_dict
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = {
    ...         "train": create_json_shard(
    ...             [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...         ),
    ...         "val": create_json_shard(
    ...             [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...         ),
    ...     }
    ...     shard = create_shard_dict(shards, uri=Path(tmpdir).joinpath("my_uri").as_uri())
    ...     shard
    ...
    ShardDict(
      (train): JsonShard(uri=file:///.../shard/uri1)
      (val): JsonShard(uri=file:///.../shard/uri2)
    )

    ```
    """
    logger.info(f"Saving URI file {uri}")
    JsonSaver().save(ShardDict.generate_uri_config(shards), sanitize_path(uri))
    return ShardDict(uri, shards)



r"""Contain some comparators to use ``BaseShard`` objects with
``coola.objects_are_equal``."""

from __future__ import annotations

__all__ = ["ShardEqualityComparator"]

import logging
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

from iden.shard.base import BaseShard

if TYPE_CHECKING:
    import sys

    from coola.equality import EqualityConfig

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

logger = logging.getLogger(__name__)


class ShardEqualityComparator(BaseEqualityComparator[BaseShard]):
    r"""Implement an equality comparator for ``BaseShard`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> Self:
        return self.__class__()

    def equal(self, actual: Any, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseShard):  # pragma: no cover
    EqualityTester.add_comparator(BaseShard, ShardEqualityComparator())



r"""Contain file-based shard implementations."""

from __future__ import annotations

__all__ = ["FileShard"]

from typing import TYPE_CHECKING, Any, TypeVar

from coola import objects_are_equal
from objectory import OBJECT_TARGET

from iden.constants import KWARGS, LOADER
from iden.io import AutoFileLoader, BaseLoader, load_json, setup_loader
from iden.shard.base import BaseShard
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

T = TypeVar("T")


class FileShard(BaseShard[T]):
    r"""Implement a generic shard where the data are stored in a single
    file.

    Args:
        uri: The shard's URI.
        path: Specifies the path to the pickle file.
        loader: The data loader or its configuration.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import FileShard
    >>> from iden.io import save_json, JsonLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     file = Path(tmpdir).joinpath("data.json")
    ...     save_json([1, 2, 3], file)
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     shard = FileShard(uri=uri, path=file, loader=JsonLoader())
    ...     shard.get_data()
    ...
    [1, 2, 3]

    ```
    """

    def __init__(
        self, uri: str, path: Path | str, loader: BaseLoader[T] | dict | None = None
    ) -> None:
        self._uri = uri
        self._path = sanitize_path(path)
        self._loader = setup_loader(loader or AutoFileLoader())

        self._is_initialized = False
        self._data = None

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(uri={self.get_uri()})"

    @property
    def path(self) -> Path:
        r"""The path to the file with data."""
        return self._path

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(
            self.get_uri(), other.get_uri(), equal_nan=equal_nan
        ) and objects_are_equal(self.path, other.path, equal_nan=equal_nan)

    def get_data(self) -> T:
        if not self._is_initialized:
            self._data = self._loader.load(self._path)
            self._is_initialized = True
        return self._data

    def get_uri(self) -> str:
        return self._uri

    @classmethod
    def from_uri(cls, uri: str) -> FileShard:
        r"""Instantiate a shard from its URI.

        Args:
            uri: The URI.

        Returns:
            The instantiated shard.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import FileShard, create_json_shard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
        ...     _ = create_json_shard([1, 2, 3], uri=uri)
        ...     shard = FileShard.from_uri(uri)
        ...     shard
        ...
        FileShard(uri=file:///.../my_uri)

        ```
        """
        config = load_json(sanitize_path(uri))
        return cls(uri=uri, **config[KWARGS])

    @classmethod
    def generate_uri_config(cls, path: Path) -> dict:
        r"""Generate the minimal config that is used to load the shard
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            path: The path to the json file.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import FileShard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     file = Path(tmpdir).joinpath("data.json")
        ...     FileShard.generate_uri_config(file)
        ...
        {'kwargs': {'path': '.../data.json'},
         'loader': {'_target_': 'iden.shard.loader.FileShardLoader'}}

        ```
        """
        return {
            KWARGS: {"path": sanitize_path(path).as_posix()},
            LOADER: {OBJECT_TARGET: "iden.shard.loader.FileShardLoader"},
        }



r"""Contain implementation of savers that use the safetensors format."""

from __future__ import annotations

__all__ = ["NumpySafetensorsSaver", "TorchSafetensorsSaver"]

from typing import TYPE_CHECKING
from unittest.mock import Mock

from coola.utils import check_numpy, check_torch, is_numpy_available, is_torch_available

from iden.io.base import BaseFileSaver
from iden.utils.imports import check_safetensors, is_safetensors_available

if TYPE_CHECKING:
    from pathlib import Path


if is_safetensors_available() and is_numpy_available():
    import numpy as np
    from safetensors import numpy as sn
else:  # pragma: no cover
    np = Mock()
    sn = Mock()

if is_safetensors_available() and is_torch_available():
    import torch
    from safetensors import torch as st
else:  # pragma: no cover
    st = Mock()
    torch = Mock()


class NumpySafetensorsSaver(BaseFileSaver[dict[str, np.ndarray]]):
    r"""Implement a file saver to save ``numpy.ndarray``s with the
    safetensors format.

    This saver can only save a dictionary of ``numpy.ndarray``s.

    Link: https://huggingface.co/docs/safetensors/en/index
    """

    def __init__(self) -> None:
        check_safetensors()
        check_numpy()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _save_file(self, to_save: dict[str, np.ndarray], path: Path) -> None:
        sn.save_file(to_save, path)


class TorchSafetensorsSaver(BaseFileSaver[dict[str, torch.Tensor]]):
    r"""Implement a file saver to save ``torch.Tensor``s with the
    safetensors format.

    This saver can only save a dictionary of ``torch.Tensor``s.

    Link: https://huggingface.co/docs/safetensors/en/index
    """

    def __init__(self) -> None:
        check_safetensors()
        check_torch()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _save_file(self, to_save: dict[str, torch.Tensor], path: Path) -> None:
        st.save_file(to_save, path)



r"""Contain the base class to implement a shard object."""

from __future__ import annotations

__all__ = ["BaseShard"]

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

T = TypeVar("T")


class BaseShard(Generic[T], ABC):
    r"""Define the base class to implement a shard."""

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two shards are equal or not.

        Args:
            other: The object to compare with.
            equal_nan: If ``True``, then two ``NaN``s will be
                considered equal.

        Returns:
            ``True`` if the two shards are equal, otherwise ``False``.
        """

    @abstractmethod
    def get_data(self) -> T:
        r"""Get the data in the shard.

        Returns:
            The data in the shard.
        """

    @abstractmethod
    def get_uri(self) -> str | None:
        r"""Get the Uniform Resource Identifier (URI) of the shard.

        Returns:
            The Uniform Resource Identifier (URI).
        """



r"""Contain in-memory shard implementations."""

from __future__ import annotations

__all__ = ["InMemoryShard"]

from typing import Any

from coola import objects_are_equal

from iden.shard.base import BaseShard


class InMemoryShard(BaseShard[Any]):
    r"""Implement an in-memory shard.

    This shard does not have valid URI as the data are stored
    in-memory.

    Example usage:

    ```pycon
    >>> from iden.shard import InMemoryShard
    >>> shard = InMemoryShard([1, 2, 3])
    >>> shard.get_data()
    [1, 2, 3]

    ```
    """

    def __init__(self, data: Any) -> None:
        self._data = data

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._data, other._data, equal_nan=equal_nan)

    def get_data(self) -> Any:
        return self._data

    def get_uri(self) -> str | None:
        return None



r"""Contain PyTorch shard loader implementations."""

from __future__ import annotations

__all__ = ["TorchShardLoader"]

from typing import Any

from coola.utils import check_torch

from iden.shard.loader.base import BaseShardLoader
from iden.shard.torch import TorchShard


class TorchShardLoader(BaseShardLoader[Any]):
    r"""Implement a PyTorch shard loader.

    Raises:
        RuntimeError: if ``torch`` is not installed.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_torch_shard
    >>> from iden.shard.loader import TorchShardLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     _ = create_torch_shard([1, 2, 3], uri=uri)
    ...     loader = TorchShardLoader()
    ...     shard = loader.load(uri)
    ...     shard
    ...
    TorchShard(uri=file:///.../my_uri)

    ```
    """

    def __init__(self) -> None:
        check_torch()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> TorchShard:
        return TorchShard.from_uri(uri)



r"""Contain JSON shard loader implementations."""

from __future__ import annotations

__all__ = ["JsonShardLoader"]

from typing import Any, TypeVar

from iden.shard.json import JsonShard
from iden.shard.loader.base import BaseShardLoader

T = TypeVar("T")


class JsonShardLoader(BaseShardLoader[Any]):
    r"""Implement a JSON shard loader.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard
    >>> from iden.shard.loader import JsonShardLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     _ = create_json_shard([1, 2, 3], uri=uri)
    ...     loader = JsonShardLoader()
    ...     shard = loader.load(uri)
    ...     shard
    ...
    JsonShard(uri=file:///.../my_uri)

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> JsonShard:
        return JsonShard.from_uri(uri)



r"""Contain safetensors-based shard implementations."""

from __future__ import annotations

__all__ = ["TorchSafetensorsShard", "create_torch_safetensors_shard"]

import logging
from typing import TYPE_CHECKING
from unittest.mock import Mock

from coola.utils import is_torch_available
from objectory import OBJECT_TARGET

from iden.constants import KWARGS, LOADER
from iden.io import JsonSaver
from iden.io.safetensors import TorchLoader, TorchSaver
from iden.shard.file import FileShard
from iden.utils.path import sanitize_path

if is_torch_available():
    import torch
else:  # pragma: no cover
    torch = Mock()

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class TorchSafetensorsShard(FileShard[dict[str, torch.Tensor]]):
    r"""Implement a safetensors shard for ``torch.Tensor``s.

    The data are stored in a safetensors file.

    Args:
        uri: The shard's URI.
        path: Specifies the path to the safetensors file.

    Raises:
        RuntimeError: if ``safetensors`` or ``torch`` is not installed.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import TorchSafetensorsShard
    >>> from iden.io.safetensors import TorchSaver
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     file = Path(tmpdir).joinpath("data.safetensors")
    ...     TorchSaver().save({"key1": torch.ones(2, 3), "key2": torch.arange(5)}, file)
    ...     shard = TorchSafetensorsShard(uri="file:///data/1234456789", path=file)
    ...     shard.get_data()
    ...
    {'key1': tensor([[1., 1., 1.], [1., 1., 1.]]), 'key2': tensor([0, 1, 2, 3, 4])}

    ```
    """

    def __init__(self, uri: str, path: Path | str) -> None:
        super().__init__(uri, path, loader=TorchLoader())

    @classmethod
    def generate_uri_config(cls, path: Path) -> dict:
        r"""Generate the minimal config that is used to load the shard
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            path: The path to the pickle file.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import TorchSafetensorsShard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     file = Path(tmpdir).joinpath("data.safetensors")
        ...     TorchSafetensorsShard.generate_uri_config(file)
        ...
        {'kwargs': {'path': '.../data.safetensors'},
         'loader': {'_target_': 'iden.shard.loader.TorchSafetensorsShardLoader'}}

        ```
        """
        return {
            KWARGS: {"path": sanitize_path(path).as_posix()},
            LOADER: {OBJECT_TARGET: "iden.shard.loader.TorchSafetensorsShardLoader"},
        }


def create_torch_safetensors_shard(
    data: dict[str, torch.Tensor], uri: str, path: Path | None = None
) -> TorchSafetensorsShard:
    r"""Create a ``TorchSafetensorsShard`` from data.

    Note:
        It is a utility function to create a ``TorchSafetensorsShard``
            from its data and URI. It is possible to create a
            ``TorchSafetensorsShard`` in other ways.

    Args:
        data: The data to save in the safetensors file.
        uri: The shard's URI.
        path: The path to the safetensors file. If ``None``, a path is
            automatically based on the URI.

    Returns:
        The ``TorchSafetensorsShard`` object.

    Raises:
        RuntimeError: if ``safetensors`` or ``torch`` is not installed.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> import torch
    >>> from iden.shard import create_torch_safetensors_shard
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shard = create_torch_safetensors_shard(
    ...         data={"key1": torch.ones(2, 3), "key2": torch.arange(5)},
    ...         uri=Path(tmpdir).joinpath("my_uri").as_uri()
    ...     )
    ...     shard.get_data()
    ...
    {'key1': tensor([[1., 1., 1.], [1., 1., 1.]]), 'key2': tensor([0, 1, 2, 3, 4])}

    ```
    """
    if path is None:
        path = sanitize_path(uri + ".safetensors")
    logger.info(f"Saving URI file {uri}")
    JsonSaver().save(TorchSafetensorsShard.generate_uri_config(path), sanitize_path(uri))
    logger.info(f"Saving data in file {path}")
    TorchSaver().save(data, path)
    return TorchSafetensorsShard(uri, path)



r"""Contain JSON-based shard implementations."""

from __future__ import annotations

__all__ = ["JsonShard", "create_json_shard"]

import logging
from typing import TYPE_CHECKING, Any

from objectory import OBJECT_TARGET

from iden.constants import KWARGS, LOADER
from iden.io import JsonLoader, JsonSaver
from iden.shard.file import FileShard
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class JsonShard(FileShard[Any]):
    r"""Implement a JSON shard.

    The data are stored in a JSON file.

    Args:
        uri: The shard's URI.
        path: Specifies the path to the JSON file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import JsonShard
    >>> from iden.io import save_json
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     file = Path(tmpdir).joinpath("data.json")
    ...     save_json([1, 2, 3], file)
    ...     shard = JsonShard(uri="file:///data/1234456789", path=file)
    ...     shard.get_data()
    ...
    [1, 2, 3]

    ```
    """

    def __init__(self, uri: str, path: Path | str) -> None:
        super().__init__(uri, path, loader=JsonLoader())

    @classmethod
    def generate_uri_config(cls, path: Path) -> dict:
        r"""Generate the minimal config that is used to load the shard
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            path: The path to the json file.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import JsonShard
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     file = Path(tmpdir).joinpath("data.json")
        ...     JsonShard.generate_uri_config(file)
        ...
        {'kwargs': {'path': '.../data.json'},
         'loader': {'_target_': 'iden.shard.loader.JsonShardLoader'}}

        ```
        """
        return {
            KWARGS: {"path": sanitize_path(path).as_posix()},
            LOADER: {OBJECT_TARGET: "iden.shard.loader.JsonShardLoader"},
        }


def create_json_shard(data: Any, uri: str, path: Path | None = None) -> JsonShard:
    r"""Create a ``JsonShard`` from data.

    Note:
        It is a utility function to create a ``JsonShard`` from its
            data and URI. It is possible to create a ``JsonShard``
            in other ways.

    Args:
        data: The data to save in the json file.
        uri: The shard's URI.
        path: The path to the JSON file. If ``None``, a path is
            automatically based on the URI.

    Returns:
        The ``JsonShard`` object.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shard = create_json_shard([1, 2, 3], uri=Path(tmpdir).joinpath("my_uri").as_uri())
    ...     shard.get_data()
    ...
    [1, 2, 3]

    ```
    """
    if path is None:
        path = sanitize_path(uri + ".json")
    logger.info(f"Saving URI file {uri}")
    JsonSaver().save(JsonShard.generate_uri_config(path), sanitize_path(uri))
    logger.info(f"Saving data in file {path}")
    JsonSaver().save(data, path)
    return JsonShard(uri, path)



r"""Contain a data loader that automatically find the right loader to
load the data based on the file extension."""

from __future__ import annotations

__all__ = ["AutoFileLoader"]

from typing import TYPE_CHECKING, Any

from coola.utils import str_indent, str_mapping

from iden.io.base import BaseLoader
from iden.io.utils import get_loader_mapping

if TYPE_CHECKING:
    from pathlib import Path


class AutoFileLoader(BaseLoader[Any]):
    r"""Implement a data loader to load data based on the file extension.

    Example usage:

    ```pycon
    >>> from pathlib import Path
    >>> from iden.io import AutoFileLoader
    >>> data = AutoFileLoader().load(Path("/path/to/data.json"))  # xdoctest: +SKIP()

    ```
    """

    registry: dict[str, BaseLoader] = {}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self.registry))}\n)"

    def load(self, path: Path) -> Any:
        extension = "".join(path.suffixes)[1:]
        loader = self.find_loader(extension)
        return loader.load(path)

    @classmethod
    def add_loader(cls, extension: str, loader: BaseLoader, exist_ok: bool = False) -> None:
        r"""Add a loader for a given file extension.

        Args:
            extension: The file extension.
            loader: The loader to use for the given file extension.
            exist_ok: If ``False``, ``RuntimeError`` is raised if the
                extension already exists. This parameter should be set
                to ``True`` to overwrite the loader for an extension.

        Raises:
            RuntimeError: if a loader is already registered for the
                extension and ``exist_ok=False``.

        Example usage:

        ```pycon
        >>> from iden.io import AutoFileLoader, TextLoader
        >>> AutoFileLoader.add_loader("text", TextLoader())

        ```
        """
        if extension in cls.registry and not exist_ok:
            msg = (
                f"A loader ({cls.registry[extension]}) is already registered for the file "
                f"extension {extension}. Please use `exist_ok=True` if you want to overwrite the "
                "loader for this extension"
            )
            raise RuntimeError(msg)
        cls.registry[extension] = loader

    @classmethod
    def has_loader(cls, extension: str) -> bool:
        r"""Indicate if a loader is registered for the given file
        extension.

        Args:
            extension: The file extension.

        Returns:
            ``True`` if a loader comparator is registered,
                otherwise ``False``.

        Example usage:

        ```pycon
        >>> from iden.io import AutoFileLoader
        >>> AutoFileLoader.has_loader("txt")
        True
        >>> AutoFileLoader.has_loader("newtxt")
        False

        ```
        """
        return extension in cls.registry

    @classmethod
    def find_loader(cls, extension: str) -> BaseLoader:
        r"""Find the loader associated to the file extension.

        Args:
            extension: The file extension.

        Returns:
            The loader for the given file extension.

        Raises:
            TypeError: if the file extension is not registered.

        Example usage:

        ```pycon
        >>> from iden.io import AutoFileLoader
        >>> AutoFileLoader.find_loader("txt")
        TextLoader()
        >>> AutoFileLoader.find_loader("json")
        JsonLoader()

        ```
        """
        if (loader := cls.registry.get(extension, None)) is not None:
            return loader
        msg = f"Incorrect extension: {extension}"
        raise TypeError(msg)


def register_auto_loaders() -> None:
    r"""Register some loaders for ``AutoFileLoader``."""
    for extension, loader in get_loader_mapping().items():
        if not AutoFileLoader.has_loader(extension):  # pragma: no cover
            AutoFileLoader.add_loader(extension, loader)



r"""Contain the base class to implement a shard loader object."""

from __future__ import annotations

__all__ = ["BaseShardLoader", "is_shard_loader_config", "setup_shard_loader"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    from iden.shard import BaseShard

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseShardLoader(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a shard loader.

    A shard loader object allows to load a ``BaseShard`` object from
    its Uniform Resource Identifier (URI).

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard import create_json_shard
    >>> from iden.shard.loader import JsonShardLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
    ...     _ = create_json_shard([1, 2, 3], uri=uri)
    ...     loader = JsonShardLoader()
    ...     loader
    ...
    JsonShardLoader()

    ```
    """

    @abstractmethod
    def load(self, uri: str) -> BaseShard[T]:
        r"""Load a shard from its Uniform Resource Identifier (URI).

        Args:
            uri: The URI of the shard to load.

        Returns:
            The loaded shard.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.shard import create_json_shard
        >>> from iden.shard.loader import JsonShardLoader
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     uri = Path(tmpdir).joinpath("my_uri").as_uri()
        ...     _ = create_json_shard([1, 2, 3], uri=uri)
        ...     loader = JsonShardLoader()
        ...     shard = loader.load(uri)
        ...     shard
        ...
        JsonShard(uri=file:///.../my_uri)

        ```
        """


def is_shard_loader_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseShardLoader``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseShardLoader`` object.

    Example usage:

    ```pycon
    >>> from iden.shard.loader import is_shard_loader_config
    >>> is_shard_loader_config({"_target_": "iden.shard.loader.PickleShardLoader"})
    True

    ```
    """
    return is_object_config(config, BaseShardLoader)


def setup_shard_loader(shard_loader: BaseShardLoader | dict) -> BaseShardLoader:
    r"""Set up a shard loader.

    The shard loader is instantiated from its configuration by using the
    ``BaseShardLoader`` factory function.

    Args:
        shard_loader: Specifies the shard loader or its configuration.

    Returns:
        The instantiated shard loader.

    Example usage:

    ```pycon
    >>> from iden.shard.loader import setup_shard_loader
    >>> shard_loader = setup_shard_loader({"_target_": "iden.shard.loader.PickleShardLoader"})
    >>> shard_loader
    PickleShardLoader()

    ```
    """
    if isinstance(shard_loader, dict):
        logger.debug("Initializing a shard loader from its configuration...")
        shard_loader = BaseShardLoader.factory(**shard_loader)
    if not isinstance(shard_loader, BaseShardLoader):
        logger.warning(f"shard loader is not a BaseShardLoader (received: {type(shard_loader)})")
    return shard_loader



r"""Implement some utility functions to manage optional dependencies."""

from __future__ import annotations

__all__ = [
    "check_safetensors",
    "check_yaml",
    "is_safetensors_available",
    "is_yaml_available",
    "safetensors_available",
    "yaml_available",
]

from importlib.util import find_spec
from typing import TYPE_CHECKING, Any

from coola.utils.imports import decorator_package_available

if TYPE_CHECKING:
    from collections.abc import Callable

#######################
#     safetensors     #
#######################


def is_safetensors_available() -> bool:
    r"""Indicate if the ``safetensors`` package is installed or not.

    Returns:
        ``True`` if ``safetensors`` is available otherwise ``False``.

    Example usage:

    ```pycon
    >>> from iden.utils.imports import is_safetensors_available
    >>> is_safetensors_available()

    ```
    """
    return find_spec("safetensors") is not None


def check_safetensors() -> None:
    r"""Check if the ``safetensors`` package is installed.

    Raises:
        RuntimeError: if the ``safetensors`` package is not installed.

    Example usage:

    ```pycon
    >>> from iden.utils.imports import check_safetensors
    >>> check_safetensors()

    ```
    """
    if not is_safetensors_available():
        msg = (
            "`safetensors` package is required but not installed. "
            "You can install `safetensors` package with the command:\n\n"
            "pip install safetensors\n"
        )
        raise RuntimeError(msg)


def safetensors_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if
    ``safetensors`` package is installed.

    Args:
        fn: Specifies the function to execute.

    Returns:
        A wrapper around ``fn`` if ``safetensors`` package is installed,
            otherwise ``None``.

    Example usage:

    ```pycon
    >>> from iden.utils.imports import safetensors_available
    >>> @safetensors_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_safetensors_available)


################
#     yaml     #
################


def is_yaml_available() -> bool:
    r"""Indicate if the ``yaml`` package is installed or not.

    Returns:
        ``True`` if ``yaml`` is available otherwise ``False``.

    Example usage:

    ```pycon
    >>> from iden.utils.imports import is_yaml_available
    >>> is_yaml_available()

    ```
    """
    return find_spec("yaml") is not None


def check_yaml() -> None:
    r"""Check if the ``yaml`` package is installed.

    Raises:
        RuntimeError: if the ``yaml`` package is not installed.

    Example usage:

    ```pycon
    >>> from iden.utils.imports import check_yaml
    >>> check_yaml()

    ```
    """
    if not is_yaml_available():
        msg = (
            "`yaml` package is required but not installed. "
            "You can install `yaml` package with the command:\n\n"
            "pip install pyyaml\n"
        )
        raise RuntimeError(msg)


def yaml_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``yaml``
    package is installed.

    Args:
        fn: Specifies the function to execute.

    Returns:
        A wrapper around ``fn`` if ``yaml`` package is installed,
            otherwise ``None``.

    Example usage:

    ```pycon
    >>> from iden.utils.imports import yaml_available
    >>> @yaml_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_yaml_available)



r"""Contain the base class to implement a dataset object."""

from __future__ import annotations

__all__ = ["VanillaDataset", "create_vanilla_dataset", "check_shards"]

import logging
from typing import TYPE_CHECKING, Any, TypeVar

from coola import objects_are_equal
from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping
from objectory import OBJECT_TARGET

from iden.constants import ASSETS, LOADER, SHARDS
from iden.dataset.base import BaseDataset
from iden.dataset.exceptions import AssetNotFoundError, SplitNotFoundError
from iden.io import JsonSaver, load_json
from iden.shard import ShardDict
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from iden.shard import BaseShard, ShardTuple

logger = logging.getLogger(__name__)

T = TypeVar("T")


class VanillaDataset(BaseDataset[T]):
    r"""Implement a simple dataset.

    Args:
        uri: The URI associated to the dataset.
        shards: The dataset's shards. Each item in the mapping
            represent a dataset split, where the key is the dataset
            split and the value is the shards.
        assets: The dataset's assets.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import VanillaDataset
    >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = create_shard_dict(
    ...         shards={
    ...             "train": create_shard_tuple(
    ...                 [
    ...                     create_json_shard(
    ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...                     ),
    ...                     create_json_shard(
    ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...                     ),
    ...                 ],
    ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
    ...             ),
    ...             "val": create_shard_tuple(
    ...                 shards=[],
    ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
    ...             ),
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
    ...     )
    ...     assets = create_shard_dict(
    ...         shards={
    ...             "stats": create_json_shard(
    ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
    ...             )
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_assets").as_uri(),
    ...     )
    ...     dataset = VanillaDataset(
    ...         uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets
    ...     )
    ...     dataset
    ...
    VanillaDataset(
      (uri): file:///.../uri
      (shards): ShardDict(
          (train): ShardTuple(
              (0): JsonShard(uri=file:///.../shard/uri1)
              (1): JsonShard(uri=file:///.../shard/uri2)
            )
          (val): ShardTuple()
        )
      (assets): ShardDict(
          (stats): JsonShard(uri=file:///.../uri_stats)
        )
    )

    ```
    """

    def __init__(
        self, uri: str, shards: ShardDict[ShardTuple[BaseShard[T]]], assets: ShardDict
    ) -> None:
        self._uri = str(uri)
        check_shards(shards)
        self._shards = shards
        self._assets = assets

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "uri": self._uri,
                    "shards": self._shards,
                    "assets": self._assets,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "uri": self._uri,
                    "shards": self._shards,
                    "assets": self._assets,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.get_uri(), other.get_uri(), equal_nan=equal_nan)
            and objects_are_equal(self._shards, other._shards, equal_nan=equal_nan)
            and objects_are_equal(self._assets, other._assets, equal_nan=equal_nan)
        )

    def get_asset(self, asset_id: str) -> BaseShard:
        if asset_id not in self._assets:
            msg = f"asset '{asset_id}' does not exist"
            raise AssetNotFoundError(msg)
        return self._assets.get_shard(asset_id)

    def has_asset(self, asset_id: str) -> bool:
        return self._assets.has_shard(asset_id)

    def get_shards(self, split: str) -> tuple[BaseShard[T], ...]:
        if split not in self._shards:
            msg = f"split '{split}' does not exist"
            raise SplitNotFoundError(msg)
        return self._shards[split].get_data()

    def get_num_shards(self, split: str) -> int:
        if split not in self._shards:
            msg = f"split '{split}' does not exist"
            raise SplitNotFoundError(msg)
        return len(self._shards.get_shard(split))

    def get_splits(self) -> set[str]:
        return self._shards.get_shard_ids()

    def has_split(self, split: str) -> bool:
        return split in self._shards

    def get_uri(self) -> str:
        return self._uri

    @classmethod
    def from_uri(cls, uri: str) -> VanillaDataset:
        r"""Instantiate a shard from its URI.

        Args:
            uri: The URI.

        Returns:
            The instantiated shard.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import create_vanilla_dataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_assets").as_uri(),
        ...     )
        ...     uri = Path(tmpdir).joinpath("uri").as_uri()
        ...     _ = create_vanilla_dataset(uri=uri, shards=shards, assets=assets)
        ...     dataset = VanillaDataset.from_uri(uri)
        ...     dataset
        ...
        VanillaDataset(
          (uri): file:///.../uri
          (shards): ShardDict(
              (train): ShardTuple(
                  (0): JsonShard(uri=file:///.../shard/uri1)
                  (1): JsonShard(uri=file:///.../shard/uri2)
                )
              (val): ShardTuple()
            )
          (assets): ShardDict(
              (stats): JsonShard(uri=file:///.../uri_stats)
            )
        )

        ```
        """
        # local import to avoid cyclic dependencies
        from iden.shard import load_from_uri

        config = load_json(sanitize_path(uri))
        shards = load_from_uri(config[SHARDS])
        assets = load_from_uri(config[ASSETS])
        return cls(uri=uri, shards=shards, assets=assets)

    @classmethod
    def generate_uri_config(
        cls,
        shards: ShardDict[ShardTuple[BaseShard[T]]],
        assets: ShardDict,
    ) -> dict:
        r"""Generate the minimal config that is used to load the dataset
        from its URI.

        The config must be compatible with the JSON format.

        Args:
            shards: The shards in the dataset. Each item in the mapping
                represent a dataset split, where the key is the dataset
                split and the value is the shards.
            assets: The dataset's assets.

        Returns:
            The minimal config to load the shard from its URI.

        Example usage:

        ```pycon

        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset import VanillaDataset
        >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     shards = create_shard_dict(
        ...         shards={
        ...             "train": create_shard_tuple(
        ...                 [
        ...                     create_json_shard(
        ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
        ...                     ),
        ...                     create_json_shard(
        ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
        ...                     ),
        ...                 ],
        ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
        ...             ),
        ...             "val": create_shard_tuple(
        ...                 shards=[],
        ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
        ...             ),
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
        ...     )
        ...     assets = create_shard_dict(
        ...         shards={
        ...             "stats": create_json_shard(
        ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
        ...             )
        ...         },
        ...         uri=Path(tmpdir).joinpath("uri_assets").as_uri(),
        ...     )
        ...     config = VanillaDataset.generate_uri_config(shards=shards, assets=assets)
        ...     config
        ...
        {'loader': {'_target_': 'iden.dataset.loader.VanillaDatasetLoader'},
         'shards': 'file:///.../uri_shards',
         'assets': 'file:///.../uri_assets'}

        ```
        """
        return {
            LOADER: {OBJECT_TARGET: "iden.dataset.loader.VanillaDatasetLoader"},
            SHARDS: shards.get_uri(),
            ASSETS: assets.get_uri(),
        }


def create_vanilla_dataset(
    shards: ShardDict[ShardTuple[BaseShard[T]]],
    assets: ShardDict,
    uri: str,
) -> VanillaDataset:
    r"""Create a ``VanillaDataset`` from its shards.

    Note:
        It is a utility function to create a ``VanillaDataset`` from
            its shards and URI. It is possible to create a
            ``VanillaDataset`` in other ways.

    Args:
        shards: The dataset's shards. Each item in the mapping
            represent a dataset split, where the key is the dataset
            split and the value is the shards.
        assets: The dataset's assets.
        uri: The URI associated to the dataset.

    Returns:
        The instantited ``VanillaDataset`` object.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import create_vanilla_dataset
    >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = create_shard_dict(
    ...         shards={
    ...             "train": create_shard_tuple(
    ...                 [
    ...                     create_json_shard(
    ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...                     ),
    ...                     create_json_shard(
    ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...                     ),
    ...                 ],
    ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
    ...             ),
    ...             "val": create_shard_tuple(
    ...                 shards=[],
    ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
    ...             ),
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
    ...     )
    ...     assets = create_shard_dict(
    ...         shards={
    ...             "stats": create_json_shard(
    ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
    ...             )
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_assets").as_uri(),
    ...     )
    ...     dataset = create_vanilla_dataset(
    ...         uri=Path(tmpdir).joinpath("uri").as_uri(), shards=shards, assets=assets
    ...     )
    ...     dataset
    ...
    VanillaDataset(
      (uri): file:///.../uri
      (shards): ShardDict(
          (train): ShardTuple(
              (0): JsonShard(uri=file:///.../shard/uri1)
              (1): JsonShard(uri=file:///.../shard/uri2)
            )
          (val): ShardTuple()
        )
      (assets): ShardDict(
          (stats): JsonShard(uri=file:///.../uri_stats)
        )
    )

    ```
    """
    logger.info(f"Saving URI file {uri}")
    JsonSaver().save(
        VanillaDataset.generate_uri_config(shards=shards, assets=assets), sanitize_path(uri)
    )
    return VanillaDataset(uri=uri, shards=shards, assets=assets)


def check_shards(shards: BaseShard) -> None:
    r"""Check if the shards have a valid configuration.

    The shards must be sorted by ascending order of URIs.

    Args:
        shards: The shards to check.

    Raises:
        TypeError: if the type is incorrect.
        RuntimeError: if the shard configuration is incorrect.
    """
    if not isinstance(shards, ShardDict):
        msg = f"Incorrect shard type: {type(shards)}"
        raise TypeError(msg)
    for shard_id in shards.get_shard_ids():
        if not shards.get_shard(shard_id).is_sorted_by_uri():
            msg = f"split '{shard_id}' is not sorted by ascending order of URIs"
            raise RuntimeError(msg)



r"""Contain data structures to manage a collection of shards."""

from __future__ import annotations

__all__ = ["ShardDict", "ShardList"]

from iden.shard.collection.dict import ShardDict
from iden.shard.collection.list import ShardList



r"""Contain the definition of the exceptions."""

from __future__ import annotations

__all__ = ["AssetExistsError", "AssetNotFoundError", "SplitNotFoundError"]


class AssetExistsError(Exception):
    r"""Raised when trying to add an asset that already exists."""


class AssetNotFoundError(Exception):
    r"""Raised when trying to access an asset that does not exist."""


class SplitNotFoundError(Exception):
    r"""Raised when trying to access a split that does not exist."""
